#include <string.h>
#include <ctype.h>
#include "tetris.h"
#include "draw.h"
#include "term.h"

char tetrom_colors[7] = {
	1,  /* I red */
	7,  /* J white */
	5,  /* L magenta */
	4,  /* O blue */
	2,  /* S green */
	3,  /* T yellow/brown */
	6   /* Z cyan */
};

short block_chars[2] = {TEXTURE1, TEXTURE2};
short bgdot = 0;

static void setblockchars_tt(int clr)
{
	int c = 0;
	switch (clr) {
	case 1:
		block_chars[0] = '<';
		block_chars[1] = '>';
		break;
	case 2:
		c = '%';
		break;
	case 3:
		c = '#';
		break;
	case 4:
		block_chars[0] = '[';
		block_chars[1] = ']';
		break;
	case 5:
		block_chars[0] = '(';
		block_chars[1] = ')';
		break;
	case 6:
		c = '@';
		break;
	case 7:
		block_chars[0] = '{';
		block_chars[1] = '}';
	}
	if (c) {
		block_chars[0] = c;
		block_chars[1] = c;
	}
}

static void setcolor(int clr)
{
	if (!clr) {
		setattr_normal();
		setcolorpair(BOARD_BG_COLOR);
		return;
	}
	if (!_MONOCHROME)
		setcolorpair(clr);
	else if (!_TT_BLOCKS) {
		setattr_normal();
		setattr_standout();
	}
	if (term_flags & (TT_BLOCKS | TT_BLOCKS_BG))
		setblockchars_tt(clr);
}

static void drawbl(int bl, int clr, int x, int y)
{
	int c, i;
	setcolor(clr%10);
	c = block_chars[clr<10 || term_flags & (TT_BLOCKS | TT_BLOCKS_BG)];
	while (bl) {
		setcurs(x, y);
		i = 1;
		do {
			if (!(bl & i))
				movefwd(2);
			else {
				putch(block_chars[0]);
				putch(c);
			}
		} while (bl & 16-(i <<= 1));
		bl >>= 4;
		y++;
	}
}

void draw_tetris_logo(int x, int y)
{
	int bl;
	drawbl(0x227, 1, x, y);
	drawbl(0x313, 4, x+7, y);
	drawbl(1, 4, x+8, y+1);
	drawbl(0x227, 6, x+12, y);
	drawbl(0x113, 7, x+19, y);
	drawbl(0x111, 2, x+24, y);
	bl = 0x326;
	if (is_outside_screen(x+33))
		bl = 0x322;
	drawbl(bl, 3, x+27, y);
	setattr_normal();
}

#ifndef board_x
int board_x(int pl, int col)
{
	int x = 2*col;
	if (pl == 2)
		return x+35;
	if (TWOPLAYER_MODE)
		return x+1;
	return x+11;
}
#endif

void next_xy(int pl, int *x, int *y)
{
	if (_HEIGHT_24L)
		*x = board_x(pl, 3);
	else if (TWOPLAYER_MODE)
		*x = (pl==1) ? -9 : board_x(2, 11);
	else {
		*x = 1;
		*y = 15;
		return;
	}
	*y = 1;
}

static void drawstr(const char *str, int n, int x, int y)
{
	char s[8];
	const char *p;
	int m;
	while (p = strchr(str, '\\')) {
		m = p-str;
		memcpy(s, str, m);
		s[m] = '\0';
		str += m+1;
		if (isdigit(p[1])) {
			str++;
			m = p[1]-'0';
		} else
			m = 1;
		do {
			printstr_acs(s, n);
			setcurs(x, ++y);
		} while (--m);
	}
	printstr_acs(str, n);
}

static void drawboard(int pl)
{
	int x = board_x(pl, 0);
	int i;
	if (_HEIGHT_24L) {
		i = x+5;
		setwcurs(0, i, 0);
		if (!_MONOCHROME) {
			setcolorpair(BOARD_FRAME_COLOR);
			setattr_bold();
		}
		drawstr("lqNk\\x Nx\\2|_N|", 8, i, 0);
		setcolorpair(BOARD_FRAME_COLOR);
		setcurs(x, 3);
		printstr("_____");
		movefwd(10);
		printstr("_____");
		i = 4;
	} else
		i = 0;
	x--;
	setwcurs(0, x, i);
	setcolorpair(BOARD_FRAME_COLOR);
	for (i = 0; i < 20; i++) {
		putch(VLINE);
		movefwd(20);
		putch(VLINE);
		if (i < 19)
			newln(x);
	}
	x++;
	if (term_height > 20 && term_height != 24) {
		newln(x);
		printstr_acs("^N", 20);
	} else {
		setcurs(0, term_height-2);
		newln(0);
	}
}

static void drawpanel_labels(const char *first, int x)
{
	const char *s = first;
	int i;
	setcolorpair(PANEL_LABEL_COLOR);
	for (i = 1; i <= 9; i += 4) {
		setcurs(x, i);
		putch(' ');
		printstr(s);
		putch(' ');
		s = s==first ? "Level" : "Lines";
	}
}

static void printstat_1p()
{
	setattr_normal();
	if (!_WHITE_BG)
		setattr_bold();
	setcurs(1, 2);
	printint(" %06d ", player1.score % 1000000);
	setcurs(3, 6);
	printint(" %02d ", player1.level);
	setcurs(3, 10);
	printint(" %03d ", player1.lines);
}

static void drawpanel_bordercolor(int clr)
{
	if (_MONOCHROME)
		setattr_normal();
	else
		setcolorpair(clr | 16);
}

static int draw_vline(int x, int y, int h)
{
	if (is_outside_screen(x+1))
		return 0;
	movefwd(x);
	while (--h) {
		putch(VLINE);
		newln(x);
	}
	putch(VLINE);
	return 1;
}

static void drawpanel_1p()
{
	int h24 = _HEIGHT_24L;
	int clr = (player1.level % 6)+1;
	int i;
	setwcurs(WIN_PANEL, 0, 0);
	setcolor(clr);
	printstr_acs("lqNu", 8);
	for (i = 0; i <= 8; i += 4)
		drawstr("\\xhNx\\3tqNu", 8, 0, i);
	if (h24) {
		setcurs(0, 12);
		putch(LOWLEFT);
		drawpanel_labels("Score", 1);
	} else {
		setcurs(0, 13);
		printstr_acs("xhNx", 8);
		drawpanel_labels("Score", 1);
		setcurs(2, 13);
		printstr(" Next ");
	}
	printstat_1p();
	drawpanel_bordercolor(clr);
	if (h24)
		drawstr("\\x\\6x", 6, 9, 12);
	else {
		drawstr("\\x Nx\\3mqNu", 8, 0, 13);
		drawstr("\\x\\x", 0, 9, 17);
	}
	i = 0;
	if (h24)
		i = 4;
	setwcurs(0, 0, i);
	drawpanel_bordercolor(clr);
	draw_vline(board_x(1, 10)+1, i, 20);
}

static void print_tetr_stats()
{
	const char letters[8] = "IJLOSTZ";
	int sum = 0;
	int i;
	if (term_width < 45)
		return;
	setwcurs(WIN_TETROM_STATS, 1, 0);
	for (i = 0; i < 7; i++) {
		setcolor(tetrom_colors[i]);
		putch(block_chars[0]);
		putch(letters[i]);
		putch(block_chars[1]);
		setattr_normal();
		printint(" %03d", tetr_stats[i]);
		newln(1);
		sum += tetr_stats[i];
	}
	printstr("  -----");
	newln(0);
	printint("Sum %04d", sum);
	refreshwin(WIN_TETROM_STATS);
}

void drawgamescreen_1p()
{
	drawboard(1);
	refreshwin(0);
	redrawboard(&player1, 19);
	drawpanel_1p();
	print_top_scores();
	print_tetr_stats();
	refreshwin(WIN_PANEL);
}

#ifdef TWOPLAYER
static void drawboard_frame_2p()
{
	int x = board_x(2, 10)+1;
	int y = _HEIGHT_24L ? 4 : 0;
	setwcurs(0, 0, y);
	if (!_MONOCHROME) {
		setcolorpair(BOARD_FRAME_COLOR);
		setattr_bold();
	}
	if (draw_vline(x, y, 20)) {
		x = board_x(1, 0)-2;
		setcurs(x, y);
		drawstr("x\\9x\\9x\\x", 0, x, y);
	}
}

static void printstat_2p(const struct player *p)
{
	int x;
	setattr_normal();
	if (!_WHITE_BG)
		setattr_bold();
	x = isplayer2(p) ? 7 : 2;
	setcurs(x, 2);
	printint(" %d ", p->score);
	if (x == 2)
		x--;
	setcurs(x, 6);
	printint(" %02d ", p->level);
	setcurs(x, 10);
	printint(" %02d ", p->lines);
}

static void drawpanel_2p()
{
	int i;
	setwcurs(WIN_PANEL, 0, 0);
	setcolor(4);
	printstr_acs("tqNu", 10);
	for (i = 0; i <= 8; i += 4)
		drawstr("\\xhNx\\3tqNu", 10, 0, i);
	drawpanel_labels("Wins", 3);
	printstat_2p(&player1);
	printstat_2p(&player2);
	drawpanel_bordercolor(4);
	drawstr("\\x Nx\\6x Nx", 10, 0, 12);
}

void drawgamescreen_2p()
{
	drawboard(1);
	drawboard(2);
	drawboard_frame_2p();
	refreshwin(0);
	redrawboard(&player1, 19);
	redrawboard(&player2, 19);
	drawpanel_2p();
	refreshwin(WIN_PANEL);
}
#else
#define printstat_2p(p)
#endif

static void printpresskey()
{
	setcolorpair(MAGENTA_FG);
	printstr("PRESS KEY");
	refreshwin(0);
}

void print_press_key()
{
	int x = board_x(1, 3) - !(player1.rotationsys & ROT_LEFTHAND);
	setwcurs(0, x, _HEIGHT_24L ? 7 : 3);
	setattr_normal();
	printpresskey();
	if (TWOPLAYER_MODE) {
		movefwd(board_x(2,3)-x-9-!(player2.rotationsys &
					   ROT_LEFTHAND));
		printpresskey();
	}
	setcurs_end();
}

void drawnext(const struct player *p, const struct tetr *next)
{
	int win = WIN_NEXT+isplayer2(p);
	int bl;
	int x;
	if (TWOPLAYER_MODE && !_HEIGHT_24L && term_width < 76)
		return;
	clearwin(win);
	if (next && (bl = next->blocks)) {
		x = (bl!=TETR_I && bl!=TETR_O);
		while (!(bl & 15))
			bl >>= 4;
		setcurs(x, 0);
		drawbl(bl, next->color+10, x, 0);
	}
	upd_screen(win);
}

static void drawboard_bg(const struct player *p, int col)
{
	int lefthand = p->rotationsys & ROT_LEFTHAND;
	putch( (lefthand && col%2) ? bgdot : ' ');
	putch(!(lefthand || col%2) ? bgdot : ' ');
}

void drawblocks(const struct player *p, int bl, int x, int y,
		enum drawmode mode)
{
	int i, j;
	x *= 2;
	while (y < 0) {
		bl >>= 4;
		y++;
	}
	setwcurs(1+isplayer2(p), x, y);
	if (mode != CLEAR_BLOCKS) {
		i = p->piece.color;
		if (mode == DRAW_PIECE)
			i += 10;
		drawbl(bl, i, x, y);
	} else {
		setcolor(0);
		while (1) {
			i = 1;
			j = x/2;
			do {
				if (bl & i)
					drawboard_bg(p, j);
				else
					movefwd(2);
				j++;
			} while (bl & 16-(i <<= 1));
			if (!(bl >>= 4))
				break;
			y++;
			setcurs(x, y);
		}
	}
}

void drawrow(const struct player *p, int y)
{
	uint_least32_t bl = p->board[y];
	int clr = -1;
	int i = 0;
	setwcurs(1+isplayer2(p), 0, y);
	do {
		if (clr != (bl & 7)) {
			clr = bl & 7;
			setcolor(clr);
		}
		if (clr) {
			putch(block_chars[0]);
			putch(block_chars[1]);
		} else
			drawboard_bg(p, i);
		bl >>= 3;
	} while (++i < 10);
}

void redrawboard(const struct player *plr, int y)
{
	while (y >= 0) {
		drawrow(plr, y);
		y--;
	}
	if (plr->piece.blocks)
		drawpiece(plr, plr->piece.blocks, plr->piece.x, plr->piece.y);
	refreshwin(1+isplayer2(plr));
}

void clearboard_paused()
{
	int x, y;
	int i, j;
	getwin_xy(1, &x, &y);
	setwcurs(0, x, y);
	setcolor(0);
	for (i = 0; i < 20; i++) {
		for (j = 0; j < 10; j++)
			drawboard_bg(&player1, j);
		if (i < 19)
			newln(x);
	}
	setcurs(x+5-!(player1.rotationsys & ROT_LEFTHAND), y+3);
	setcolorpair(MAGENTA_FG);
	printstr("-- PAUSE --");
	refreshwin(0);
}

void upd_stat(const struct player *p, int levelup)
{
	setwcurs(WIN_PANEL, 1, 2);
	if (TWOPLAYER_MODE)
		printstat_2p(p);
	else {
		if (levelup)
			drawpanel_1p();
		else
			printstat_1p();
		print_top_scores();
		print_tetr_stats();
	}
	refreshwin(WIN_PANEL);
}

#ifdef TWOPLAYER
void upd_garbagemeter(const struct player *p, int n)
{
	int x = isplayer2(p) ? 10  : 1;
	int i = 7;
	int c = ' ';
	setwcurs(WIN_PANEL, x, 12);
	while (i) {
		if (i <= n) {
			c = UPARROW;
			setcolor(1);
			n = 0;
		}
		newln(x);
		putch(c);
		i--;
	}
	refreshwin(WIN_PANEL);
}

void show_winner(const struct player *p)
{
	int i = 1;
	while (1) {
		setwcurs(i, 4, 2);
		setattr_normal();
		setcolorpair(MAGENTA_FG);
		drawbox(4, 2, 12, 3, NULL);
		setcurs(6, 3);
		printstr("YOU ");
		printstr(p==game->player+(i==2) ? "WIN!" : "LOSE");
		refreshwin(i);
		if (i == 2)
			break;
		i = 2;
	}
	setcurs_end();
}
#endif /* TWOPLAYER */

void drawbox(int x, int y, int w, int h, const char *title)
{
	int i;
	setcurs(x, y);
	if (!title)
		printstr_acs("lqNk", w-2);
	else {
		setattr_normal();
		setattr_standout();
		i = w-strlen(title)-4;
		printstr_acs("lqN ", i/2);
		printstr(title);
		printstr_acs(" qNk", i-i/2);
		setattr_normal();
	}
	for (i = 2; i < h; i++) {
		newln(x);
		printstr_acs("x Nx", w-2);
	}
	newln(x);
	printstr_acs("mqNj", w-2);
}

void clearbox(int x, int y, int w, int h)
{
	setattr_normal();
	setcurs(x, y);
	while (1) {
		if (!w)
			cleartoeol();
		else
			putnchars(' ', w);
		h--;
		if (!h)
			break;
		newln(x);
	}
}

#ifdef TWOPLAYER
void draw_2p_menu_decor(int pl, int x, int y)
{
	x++;
	drawstr("\\l\\x", 0, x, y-1);
	y += 2;
	setcurs(x-1, y);
	putch(pl+'0');
	drawstr("P\\x\\2m", 0, x, y);
}
#endif

void upd_screen(int i)
{
	refreshwin(i);
	setcurs_end();
}
