#include <stdlib.h>
#ifdef HAS_ALLOCA_H
#include <alloca.h>
#endif
#include <string.h>
#ifndef TWOPLAYER
#define TWOPLAYER
#endif
#include "tetris.h"
#include "input.h"
#include "draw.h"
#include "socket.h"

#ifdef SOCKET
#define NETPLAY (game->mode & MODE_NETWORK)
#else
#define NETPLAY 0
#endif

struct times {
	int falltm;
	int delay;
};

static char *clearedlines_2p;
static char *garbage;

static struct lockdelay_state {
	char x1, x2;
	int b1, b2;
} *lockdelay_state;

char *tetrom_seq = NULL;	/* used in socket.c */
static int tetrom_seq_len = -1;
static int tetrom_seq_pos[2];

static void free_tetrom_seq()
{
	if (tetrom_seq) {
		free(tetrom_seq);
		tetrom_seq = NULL;
		tetrom_seq_len = 0;
		tetrom_seq_pos[0] = 0;
		tetrom_seq_pos[1] = 0;
	}
}

static void gen_tetrom_seq(int i, int n)
{
	n += i;
	if (n > tetrom_seq_len) {
		tetrom_seq = !tetrom_seq ? malloc(n)
					 : realloc(tetrom_seq, n);
		if (tetrom_seq_len == -1)
			atexit(free_tetrom_seq);
		tetrom_seq_len = n;
	}
	for (; i < n; i++)
		tetrom_seq[i] = randnum(7);
}

static int num_clearedlines(struct player *plr)
{
	char *lines = clearedlines_2p+4*(plr==&player2);
	return (lines[3]) ? 4 : strlen(lines);
}

static int draw_lineclear_step(struct player *plr, struct times *times)
{
	static char px[2];
	int i = isplayer2(plr);
	char *lines = clearedlines_2p+4*i;
	int n = num_clearedlines(plr);
	int x;
	if (!px[i])
		px[i] = 6;
	x = --px[i];
	if (x) {
		for (i = 0; i < n; i++) {
			clearblocks(plr, 1, x-1, lines[i]-1);
			clearblocks(plr, 1, 10-x, lines[i]-1);
		}
		upd_screen(1+isplayer2(plr));
		times->delay = CLEAR_DELAY/5;
	} else {
		redrawboard(plr, lines[n-1]-1);
		memset(lines, 0, 4);
	}
	return x;
}

static void sendgarbage(struct player *plr)
{
	int n = num_clearedlines(plr);
	int i, j;
	if (n < 4)
		n--;
	if (n) {
		i = 3*(j = !isplayer2(plr));
		garbage[i] += n;
		upd_garbagemeter(game->player+j, garbage[i]);
	}
}

static void getgarbage(struct player *plr, char *garbg)
{
	int i, clr;
	uint_least32_t bl;
	if (!garbg[0])
		return;
	while (garbg[0] && !plr->board[0]) {
		if (!garbg[1] || garbg[2] == 8) {
			do i = randnum(10)+1;
			while (i == garbg[1]);
			garbg[1] = i;
			garbg[2] = 0;
		}
		garbg[2]++;
		for (i = 0; i < 19; i++)
			plr->board[i] = plr->board[i+1];
		if (NETPLAY && isplayer2(plr))
			plr->board[19] = 0;
		else {
			bl = plr->board[18];
			while (!(bl & 7))
				bl >>= 3;
			clr = (bl & 7);
			plr->board[19] = 0;
			for (i = 0; i < 10; i++)
				if (i+1 != garbg[1])
					plr->board[19] |= clr << 3*i;
		}
		redrawboard(plr, 19);
		garbg[0]--;
	}
	upd_garbagemeter(plr, garbg[0]);
	if (garbg[0])
		plr->piece.y = -1;
	if (NETPLAY && isplayer2(plr))
		sock_requestgarbage();
}

static void spawndelay(struct player *plr, struct times *times)
{
	times->falltm = plr->falltime;
	times->delay = SPAWN_DELAY;
	getgarbage(plr, garbage+3*isplayer2(plr));
}

#ifdef SOCKET
static void sendnext(struct player *p)
{
	int i = isplayer2(p);
	if (is_server) {
		i = tetrom_seq_pos[i]+1;
		if (i >= tetrom_seq_len)
			gen_tetrom_seq(i, 1);
		sock_sendnext(p, tetrom_seq[i]);
	}
}
#else
#define sendnext(p)
#endif

static void lockdelay(struct player *plr, struct times *times)
{
	struct lockdelay_state *st = lockdelay_state+isplayer2(plr);
	struct tetr *p = &plr->piece;
	times->falltm = 0;
	times->delay = LOCK_DELAY;
	st->x1 = p->x;
	st->x2 = p->x;
	st->b1 = p->blocks;
	st->b2 = p->blocks;
	if (NETPLAY)
		sendnext(plr);
}

static int upd_lockdelay_state(struct player *plr, struct times *times)
{
	struct lockdelay_state *st = lockdelay_state+isplayer2(plr);
	struct tetr *p = &plr->piece;
	if (!hitbtm(p, plr)) {
		times->delay = 0;
		times->falltm = plr->falltime;
		st->b1 = 0;
		movedown(plr, 0);
		return 1;
	}
	if (NETPLAY && isplayer2(plr)) {
		times->delay = LOCK_DELAY;
		return 1;
	}
	if (times->delay <= 0)
		return 0;
	if (p->x != st->x2) {
		if (p->x == st->x1)
			return 0;
		times->delay = LOCK_DELAY;
		st->x1 = st->x2;
		st->x2 = p->x;
	} else if (p->blocks != st->b2) {
		if (p->blocks == st->b1)
			return 0;
		times->delay = LOCK_DELAY;
		st->b2 = p->blocks;
	}
	return 1;
}

static int lockpiece_2p(struct player *plr, struct times *times)
{
	int i = isplayer2(plr);
	if (NETPLAY) {
		if (!lockdelay_state[i].b1)
			sendnext(plr);
		if (!i && plr->piece.blocks) {
			sock_sendpiece(plr);
			sock_sendbyte(MVDOWN);
		}
	}
	lockpiece(plr);
	if (plr->piece.y < 0)
		return 0;
	if (!clearedlines[0]) {
		upd_screen(1+i);
		spawndelay(plr, times);
	} else {
		times->falltm = 0;
		memcpy(clearedlines_2p+4*i, clearedlines, 4);
		draw_lineclear_step(plr, times);
		sendgarbage(plr);
	}
	lockdelay_state[i].b1 = 0;
	return 1;
}

static void show_next(struct player *p, int i)
{
	struct tetr next;
	gettetrom(&next, i);
	drawnext(p, &next);
}

static int nextpiece(struct player *plr, struct times *times)
{
	int *i = tetrom_seq_pos;
	int *j = tetrom_seq_pos;
	char *p;

	if (plr->piece.y < 0)
		return 0;
	if (plr == &player1)
		j++;
	else {
		i++;
		times++;
	}
	if (*i+1 >= tetrom_seq_len) {
		memmove(tetrom_seq, tetrom_seq+*j, *i-*j+1);
		*i -= *j;
		*j = 0;
		gen_tetrom_seq(*i+1, 8);
	}
	p = tetrom_seq+*i;
	if (!NETPLAY || is_server)
		*i += 1;
#ifdef SOCKET
	else if (*p < 0) {
		sock_flags &= ~PL2_IN_GAME;
		return 0;
	}
#endif
	gettetrom(&plr->piece, *p);
	if (NETPLAY && p[1] < 0)
		drawnext(plr, NULL);
	else
		show_next(plr, p[1]);
	if (NETPLAY && !is_server) {
		*p = p[1];
		p[1] = -1;
	}
	times->falltm = plr->falltime;
	times->delay = 0;
	return movedown(plr, 0) && movedown(plr, 0);
}

static int getwaittm(struct times *times)
{
	int tm = (times[0].delay) ? times[0].delay : times[0].falltm;
	if (times[1].delay) {
		if (tm > times[1].delay)
			tm = times[1].delay;
	} else if (tm > times[1].falltm)
		tm = times[1].falltm;
	return tm;
}

static int processkey(struct player **plr, struct times *times)
{
	int key = getkeypress(getwaittm(times), IN_GAME);
	int flags;
	if (!key)
		return -1;
	if (key & PLAYER_2) {
		*plr = &player2;
		flags = PLAYER_2;
		times++;
	} else {
		*plr = &player1;
		flags = 0;
	}
	if (times->delay && (times->falltm ||
			     clearedlines_2p[4*((key & PLAYER_2)!=0)]))
		flags |= DISCARD_MOVES;
	return processkey_ingame(key, flags);
}

static int upd_player(struct player *plr, struct times *times, int tm)
{
	if (isplayer2(plr))
		times++;
	if (times->delay) {
		times->delay -= tm;
		if (times->falltm)
			return times->delay > 0;
		if (lockdelay_state[isplayer2(plr)].b1) {
			if (!upd_lockdelay_state(plr, times))
				return lockpiece_2p(plr, times);
		} else if (times->delay <= 0) {
			if (!draw_lineclear_step(plr, times))
				spawndelay(plr, times);
		}
	} else {
		times->falltm -= tm;
		if (times->falltm <= 0) {
			if (movedown(plr, 0))
				times->falltm = plr->falltime;
			else
				lockdelay(plr, times);
		}
	}
	return 1;
}

static int startgame_wait_2p()
{
	struct tetr next = {0};
	gettetrom(&next, *tetrom_seq);
	game->next = &next;
	return startgame_wait(0);
}

static struct player *play_round()
{
	struct player *plr;
	struct times times[2];
	int t, ret;

	memset(clearedlines_2p, 0, 8);
	memset(garbage, 0, 6);
	lockdelay_state[0].b1 = 0;
	lockdelay_state[1].b1 = 0;

	if (!NETPLAY || is_server) {
		gen_tetrom_seq(0, 8);
		show_next(&player1, *tetrom_seq);
		show_next(&player2, *tetrom_seq);
	} else {
		memset(tetrom_seq = malloc(4), -1, 4);
		tetrom_seq_len = 4;
		tetrom_seq_pos[1] = 2;
	}
	if (NETPLAY)
		sock_initround();
	if (!startgame_wait_2p()) {
		free_tetrom_seq();
		return NULL;
	}
	if (NETPLAY)
		sock_sendbyte('.');
	nextpiece(&player1, times);
	nextpiece(&player2, times);

	while (1) {
		t = gettm(0);
		ret = processkey(&plr, times);
		if (!ret && !lockpiece_2p(plr, times+isplayer2(plr)) ||
		    ret == -2 || NETPLAY && !(sock_flags & PL2_IN_GAME))
			break;
		if (ret == 2) {
			times->falltm = plr->falltime;
			t = 0;
		} else {
			t = gettm(t)-t;
			if (ret == -1)
				plr = &player1;
		}
		if (ret && !upd_player(plr, times, t) &&
		    !nextpiece(plr, times))
			break;
		plr = game->player+!isplayer2(plr);
		if (!upd_player(plr, times, t) && !nextpiece(plr, times))
			break;
	}
	free_tetrom_seq();
	if (ret == -2)
		return NULL;
	if (NETPLAY && !(sock_flags & PL2_IN_GAME)) {
		sock_sendbyte(ESC);
		plr = &player1;
	} else if (!(game->mode & MODE_BTYPE) || plr->lines)
		plr = game->player+!isplayer2(plr);
	return plr;
}

static int wait_gameover()
{
	int tm = 2000;
	int t;
	game->state = GAME_OVER;
	do {
		t = gettm(0);
		switch (getkeypress(tm, IN_GAME) & 0x7F) {
		case ESC:	return 0;
		case STARTBTN:	return 1;
		}
		tm -= gettm(t)-t;
	} while (tm > 0);
	return processkey_ingame(getkeypress_block(0), DISCARD_MOVES) != -2;
}

int startgame_2p()
{
	struct player *plr;
	player1.score = 0;
	player2.score = 0;
	drawgamescreen_2p();
	if (NETPLAY) {
		if (!sock_wait_pl2ingame())
			return 0;
		plr = &player2;
		if (plr->level != plr->startlevel) {
			setupplayer(plr);
			upd_stat(plr, 0);
		}
	}
	clearedlines_2p = alloca(8);
	garbage = alloca(6);
	lockdelay_state = alloca(2*sizeof(struct lockdelay_state));
	while (plr = play_round()) {
		plr->score++;
		upd_stat(plr, 0);
		show_winner(plr);
		if (!wait_gameover() || plr->score == 3)
			break;
		plr = &player1;
		while (1) {
			setupplayer(plr);
			redrawboard(plr, 19);
			upd_stat(plr, 0);
			upd_garbagemeter(plr, 0);
			if (isplayer2(plr))
				break;
			plr++;
		}
		if (NETPLAY) {
			if (!(sock_flags & PL2_IN_GAME)) {
				game->state = 0;
				return startgame_2p();
			}
			sock_sendbyte('s');
		}
	}
	game->state = 0;
#ifdef SOCKET
	if (NETPLAY) {
		sock_flags &= ~PL2_IN_GAME;
		sock_sendbyte(ESC);
	}
#endif
	return plr && plr->score == 3;
}
