#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdint.h>
#include <linux/input.h>
#include <linux/joystick.h>
#ifndef JOYSTICK
#define JOYSTICK
#endif
#include "joystick.h"
#include "input.h"

char js_devname[2][16] = {""};

static struct js {
	char menu_keys[7];
	char ingame_keys[5];
	uint32_t pressedbtns;
	short pressedbtns_tm[4];
	short axis_tm[2];
	char axis;
	char a_btn;
	char b_btn;
	char startbtn;
} joysticks[2];

int js_open(int i, const char *devname)
{
	char s[16] = "/dev/";
	int fd;
	if (inputdevs_fd[i+1] > 0)
		return 0;
	if (devname[0] != '/') {
		strncpy(s+5, devname, 10);
		devname = s;
	}
       	fd = open(devname, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "Could not open ");
		perror(devname);
		return 0;
	}
	inputdevs_fd[i+1] = fd;
	strncpy(js_devname[i], devname, 15);
	return 1;
}

#ifndef JSIOCGBTNMAP
#define getbtnmap(fd, js)
#else
static void getbtnmap(int fd, struct js *js)
{
	unsigned char n;
	uint16_t btnmap[KEY_MAX-BTN_MISC+1];

	if (ioctl(fd, JSIOCGBUTTONS, &n) == -1 || 
	    ioctl(fd, JSIOCGBTNMAP, btnmap) == -1 ||
	    btnmap[0] == BTN_JOYSTICK)
		return;
	while (n) {
		switch (btnmap[--n]) {
		case BTN_A:
			js->a_btn = n;
			break;
		case BTN_B:
			js->b_btn = n;
			break;
		case BTN_START:
			js->startbtn = n;
		}
	}
}
#endif

static int initjs(int fd, int i)
{
	struct js_event e;
	char name[32] = "";
	struct js *js = joysticks+i;
	int v;
	if (read(fd, &e, sizeof(struct js_event)) < sizeof(struct js_event) ||
	    !(e.type & JS_EVENT_INIT) || ioctl(fd, JSIOCGVERSION, &v) == -1)
		return 0;
	inputdevs_fd[i+1] = fd;
	js->a_btn = -1;
	js->b_btn = -1;
	ioctl(fd, JSIOCGNAME(32), name);
	if (!strcmp(name, "Smartjoy Family Super Smartjoy"))
		js->startbtn = 11;
	else
		js->startbtn = -1;
	getbtnmap(fd, js);
	return 1;
}

static int openjs(const char *devname, int i)
{
	int fd = open(devname, O_RDONLY);
	if (fd > -1) {
		if (initjs(fd, i))
			return 1;
		fprintf(stderr, "Error reading joystick device %s\n", devname);
		close(fd);
	}
	return 0;
}

static void try_openjs()
{
	char devnames[24] = "/dev/js0\0/dev/input/js0";
	char *p;
	int i = 0;
	while (1) {
		if (inputdevs_fd[i+1] == -1) {
			p = devnames;
loop:			if (openjs(p, i))
				strcpy(js_devname[i], p);
			else if (p == devnames) {
				p += 9;
				goto loop;
			}
		}
		if (i)
			break;
		i = 1;
		devnames[7] = '1';
		devnames[22] = '1';
	}
}

void init_joysticks()
{
	int *fd;
	if (*js_devname[0] && inputdevs_fd[1] == -1)
		openjs(js_devname[0], 0);
	if (*js_devname[1] && inputdevs_fd[2] == -1)
		openjs(js_devname[1], 1);
	try_openjs();
	if (*js_devname[0] && !strcmp(js_devname[0], js_devname[1])) {
		if (strchr(js_devname[0], '0')) {
			*js_devname[1] = '\0';
			fd = inputdevs_fd+2;
		} else {
			*js_devname[0] = '\0';
			fd = inputdevs_fd+1;
		}
		if (*fd > -1) {
			close(*fd);
			*fd = -1;
		}
	}
}

static void releasebtn(struct js *js, int b)
{
	uint32_t btns = js->pressedbtns;
	short *tm = js->pressedbtns_tm;
	int i = 0;
	uint32_t mask;
	do {
		if (b == (btns & 0x7F)) {
			memmove(tm+i, tm+i+1, (3-i)*sizeof(short));
			js->pressedbtns &= ~(0x7F << 8*i);
			if (!i)
				js->pressedbtns >>= 8;
			else {
				i *= 8;
				mask = ~0 << i;
				js->pressedbtns &= ~mask;
				js->pressedbtns |= (btns & ~0x7F) << i-8;
			}
		}
		i++;
	} while (btns >>= 8);
}

static void pressbtn(struct js *js, int b)
{
	short *tm = js->pressedbtns_tm;
	releasebtn(js, b);
	js->pressedbtns = b | js->pressedbtns << 8;
	tm[3] = tm[2];
	tm[2] = tm[1];
	tm[1] = tm[0];
	tm[0] = gettm(0);
}

static int readjs_axis(struct js_event *e, struct js *js)
{
	int axis;
	int i = 0;
	int b = 0;
	switch (e->number) {
	case 1:
		i = 2;
	case 0:
		if (e->value) {
			js->axis_tm[i>>1] = gettm(0);
			axis = 1+(e->value > 0);
			b = MVLEFT+i+axis-1;
		} else
			axis = 0;
		js->axis = (axis<<i | js->axis & 12>>i);
	}
	return b;
}

static int readjs(int fd, struct js *js)
{
	struct js_event e;
	int b, b2;
	if (read(fd, &e, sizeof(struct js_event)) < sizeof(struct js_event))
		return 0;
	b = 0;
	if (e.type & JS_EVENT_AXIS)
		b = readjs_axis(&e, js);
	else if (e.type & JS_EVENT_BUTTON) {
		if (e.number == js->a_btn)
			b = A_BTN;
		else if (e.number == js->b_btn)
			b = B_BTN;
		else if (e.number == js->startbtn)
			b = STARTBTN;
		else
			b = e.number+'0';
		if (!e.value) {
			releasebtn(js, b);
			b = 0;
		}
	} else {
		close_inputdev(1+(js!=joysticks));
		return -1;
	}
	if (waitinput(fd, 0)) {
		b2 = readjs(fd, js);
		if (b > b2 && b2)
			return b2;
	}
	return b;
}

int js_readbtn(int fd, int i)
{
	i = readjs(fd, joysticks+i);
	if (i == -1)
		return 0;
	return i;
}

static const char *getbtn(int b, const char *p, int n)
{
	while (n) {
		if (*p == b)
			return p;
		p++;
		n--;
	}
	return NULL;
}

static char *getbtnfor(struct js *js, int keypress)
{
	char *keys;
	int i;
	if (keypress & IN_GAME) {
		keys = js->ingame_keys;
		if ((char) keypress == HARDDROP)
			i = 4;
		else
			i = (keypress & 7)-MVUP;
	} else {
		keys = js->menu_keys;
		if ((char) keypress == STARTBTN)
			i = 6;
		else
			i = (keypress & 7)-MVLEFT;
	}
	return keys+i;
}

static int transl_btn(struct js *js, int b, int flags)
{
	const char *keys, *p;
	if (flags & IN_GAME) {
		keys = js->ingame_keys;
		if (p = getbtn(b, keys, 5))
			return p==keys+4 ? HARDDROP : MVUP+p-keys;
	}
	keys = js->menu_keys;
	if (p = getbtn(b, keys, 7))
		return p==keys+6 ? STARTBTN : MVLEFT+p-keys;
	return 0;
}

static int test_up_rotate(struct js *js, int b, int keypr)
{
	if (b==MVUP && keypr==MVUP) {
		b = *getbtnfor(js, MVUP | IN_GAME);
		if (b && b != MVUP)
			return 0;
	}
	return 1;
}

int js_getpress(int fd, int i, int flags)
{
	struct js *js = joysticks+i;
	int b, ret;
	b = readjs(fd, js);
	if (b == -1)
		return 0;
	if (!b)
		return b;
	ret = transl_btn(js, b & 0x7F, flags);
	if (!ret) {
		if (b == STARTBTN)
			return b;
		if (b > B_BTN)
			goto next;
		ret = b;
	}
	if (b > MVDOWN && ret != HARDDROP)
		pressbtn(js, b);
	else if (flags & IN_GAME && !test_up_rotate(js, b, ret))
		goto next;
	return ret;
next:	if (waitinput(fd, 0))
		return js_getpress(fd, i, flags);
	return 0;
}

int js_pressed(int i)
{
	struct js *js = joysticks+i;
	return js->axis || js->pressedbtns;
}

static int getautorep_btn(struct js *js, short **tm)
{
	uint32_t btns = js->pressedbtns;
	int i = 0;
	do {
		if (test_autorep_tm(js->pressedbtns_tm+i)) {
			*tm = js->pressedbtns_tm+i;
			return btns & 0x7F;
		}
		i++;
	} while (btns >>= 8);
	return 0;
}

int js_getautorepeat(int i, int flags)
{
	struct js *js = joysticks+i;
	short *tm = js->axis_tm;
	int b = 0;
	if (js->axis & 3 && test_autorep_tm(tm))
		b = MVLEFT+(js->axis & 3)-1;
	else if (js->axis & 12 && test_autorep_tm(tm+1)) {
		b = MVUP+(js->axis>>2)-1;
		tm++;
	} else if (js->pressedbtns)
		b = getautorep_btn(js, &tm);
	if (!b)
		return 0;
	i = transl_btn(js, b, flags);
	if (!i && b <= B_BTN)
		i = b;
	if (flags & IN_GAME) {
		if (!test_up_rotate(js, b, i))
			return 0;
		if (i==MVDOWN)
			*tm -= DAS_DELAY/2;
	}
	return i;
}

void js_reset_autorepeat(int i)
{
	struct js *js = joysticks+i;
	int t = gettm(0);
	js->axis_tm[0] = t;
	js->axis_tm[1] = t;
	js->pressedbtns_tm[0] = t;
	js->pressedbtns &= 0x7F;
}

static void rmbtn(struct js *js, int b, int keypress)
{
	int flags = keypress & IN_GAME;
	int old = transl_btn(js, b, flags);
	if (old && old != (char) keypress) {
		if (flags) {
			old |= flags;
			if (b != *getbtnfor(js, old))
				return;
		}
		js_rmmapping(js!=joysticks, old);
	}
}

void js_setmapping(int i, int btn, int keypress)
{
	struct js *js = joysticks+i;
	char *b;
	rmbtn(js, btn, keypress);
	b = getbtnfor(js, keypress);
	if (*b && !(keypress & IN_GAME) && (char) keypress >= MVUP) {
		keypress |= IN_GAME;
		if (*b == *getbtnfor(js, keypress))
			js_setmapping(i, btn, keypress);
	}
	*b = btn;
}

int js_getbtnfor(int i, int keypress)
{
	struct js *js = joysticks+i;
	char b = *getbtnfor(js, keypress);
	char k;
	if (b)
		return b;
	k = keypress;
	if (k == HARDDROP)
		return 0;
	if (keypress & IN_GAME) {
		b = js_getbtnfor(i, keypress ^ IN_GAME);
		if (!b || getbtn(b, js->ingame_keys, 5))
			return 0;
		return b;
	}
	switch (k) {
	case A_BTN:
	case B_BTN:
		if (!getbtn(k, js->menu_keys, 7) &&
		    (&js->a_btn)[k-A_BTN] > -1)
			break;
	case STARTBTN:
		return 0;
	}
	return k;
}

void js_rmmapping(int i, int keypress)
{
	char *b = getbtnfor(joysticks+i, keypress);
	if (*b)
		*b = 0;
}
