#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "menu.h"
#include "input.h"
#include "term.h"
#include "draw.h"
#include "lang.h"

#include "options.h"
#include "tetris.h"
#include "hiscore.h"

#ifndef NO_MENU
int in_menu = 1;
#endif

#ifdef TWOPLAYER
static int player_;
#else
#define player_ 0
#endif

static_ifnomenu void printmenuitem(const char *name, int sel)
{
#ifndef NO_MENU
	if (*name == '-') {
		putch(' ');
		if (!name[1])
			putnchars('-', 16);
		else
			printstr(name);
		return;
	}
#endif
	if (!sel)
		setcolorpair(MAGENTA_FG);
	else if (!_MONOCHROME)
		setcolorpair(WHITE_ON_BLUE);
	else
		setattr_standout();
	if (sel && (term_flags & TT_MONO)==TT_BLOCKS) {
		while (*name==' ') {
			putch(' ');
			name++;
		}
		putch('*');
	} else
		putch(' ');
	printstr(name);
	putch(' ');
	setattr_normal();
	movefwd(1);
}

static_ifnomenu void printmenuitem_options(const char *str, int sel)
{
	char buf[16];
	const char *p = str;
	int ret;
	while (1) {
		if (p = strchr(str, ' ')) {
			memcpy(buf, str, p-str);
			buf[p-str] = '\0';
		} else
			strcpy(buf, str);
		if (!sel) {
			setattr_standout();
			ret = printstr(buf);
			setattr_normal();
		} else
			ret = printstr(buf);
		if (!ret || !p)
			break;
		putch(' ');
		str = p+1;
		sel--;
	}
}

static_ifnomenu
void drawmenu(const char **menu, int n, int pos, int x, int y,
	      menuhandler *handlers)
{
	int i;
	setcurs(x, y);
	for (i = 0; i < n; i++) {
		printmenuitem(menu[i], (i==pos));
		if (handlers && handlers[i])
			handlers[i](0, &i);
		newln(x);
	}
}

static_ifnomenu
int handle_menuitem(const char **menu, int n, int *i, int x, int y,
		    menuhandler *handlers, int keypr)
{
	int j = *i;
	int m;
	keypr &= 0xFF;
	if (keypr == ESC)
		return 0;
	printmenuitem(menu[j], 1);
	if (handlers && handlers[j])
		switch (m = handlers[j](keypr, i)) {
		case 0:   break;
		case ESC: return 0;
		default:  return m;
		}
	switch (keypr) {
	case STARTBTN:
	case A_BTN:
		return 2;
	case B_BTN:
	case '\b':
		return 0;
	case 'q':
		exit(0);
	case MVUP:
		if (!j)
			return 1;
#ifdef NO_MENU
		m = -1;
#else
		m = -1-(*menu[j-1]=='-');
#endif
		break;
	case '\t':
		if (j == n-1) {
			m = -j;
			break;
		}
	case MVDOWN:
		if (j < n-1) {
#ifdef NO_MENU
			m = 1;
#else
			m = 1+(*menu[j+1]=='-');
#endif
			break;
		}
	default:
		return 1;
	}
	setcurs(x, y);
	printmenuitem(menu[j], 0);
	setcurs(x, y+m);
	printmenuitem(menu[*i+=m], 1);
	return 1;
}

static int openmenu(const char **menu, int n, int i, int x, int y,
		    menuhandler *handlers)
{
	drawmenu(menu, n, i, x, y, handlers);
	while (1) {
		setcurs(x, y+i);
		refreshwin(-1);		/* refresh active window */

		switch (handle_menuitem(menu, n, &i, x, y+i, handlers,
					getkeypress_block(SINGLE_PL))) {
		case 0:
			return 0;
		case 2:
			return i+1;
#ifndef NO_MENU
		case 3:
			return openmenu(menu, n, i, x, y, handlers);
#endif
		}
	}
}

static_ifnomenu
int handle_menuitem_2cols(const char **menu, int n, int *i, int h,
			  int x, int y, menuhandler *handlers, int keypr)
{
	int j = *i;
	int m;
	keypr &= 0xFF;
	switch (keypr) {
	case MVUP:
		if (j == h) {
			m = -1;
			goto col;
		}
		break;
	case '\t':
		if (j == n-1) {
			m = -j;
			goto col;
		}
	case MVDOWN:
		if (j == h-1) {
			m = 1;
			goto col;
		}
	}
	if (j < h) {
		if (keypr == MVRIGHT && j+h < n) {
			m = h;
			goto col;
		}
		m = handle_menuitem(menu, h, i, x, y, handlers, keypr);
	} else {
		if (keypr == MVLEFT && j-h >= 0) {
			m = -h;
			goto col;
		}
		j = *i-h;
		m = handle_menuitem(menu+h, n-h, &j, x, y, handlers, keypr);
		*i = j+h;
	}
	return m;
col:	printmenuitem(menu[j], 0);
	*i += m;
	return 3;
}

#ifndef NO_MENU
static int getdropdownwidth(const char **items, int n)
{
	int w = strlen(items[0]);
	int i = 1;
	while (i < n) {
		if (w < strlen(items[i]))
			w = strlen(items[i]);
		i++;
	}
	return w;
}

static void printdropdownitem(const char *name, int sel, int w)
{
	if (sel) {
		while (name[0]==' ' && name[1]==' ') {
			putch(' ');
			name++;
			w--;
		}
		setattr_normal();
	}
	printstr(name);
	if (strlen(name) < w)
		putch(' ');
	if (sel)
		setattr_standout();
}

int dropdownlist(const char **items, int n, int i, int x, int y)
{
	int w = getdropdownwidth(items, n);
	int j;
	setattr_standout();
	drawbox(x, y, w+2, n+2, NULL);
	x++;
	y++;
	setcurs(x, y);
	for (j = 0; j < n; j++) {
		printdropdownitem(items[j], i==j, w);
		newln(x);
	}
	while (1) {
		setcurs(x-1, y+i);
		refreshwin(-1);
		switch (getkeypress_block(SINGLE_PL) & 0xFF) {
		case STARTBTN:
		case A_BTN:
			goto out;
		case 'q':
			exit(0);
		case MVUP:
			if (!i)
				continue;
			j = -1;
			break;
		case '\t':
			if (i == n-1) {
				j = -i;
				break;
			}
		case MVDOWN:
			if (i < n-1) {
				j = 1;
				break;
			}
		case MVRIGHT:
			continue;
		default:
			i = -1;
			goto out;
		}
		setcurs(x, y+i);
		printdropdownitem(items[i], 0, w);
		i += j;
		setcurs(x, y+i);
		printdropdownitem(items[i], 1, w);
	}
out:	clearbox(x-1, y-1, w+2, n+2);
	return i+1;
}

int selectitem(const char **items, int n, int *i, int k)
{
	int x, y;
	switch (k) {
	case 0:
		break;
	case MVLEFT:
		if (*i)
			*i -= 1;
		break;
	case MVRIGHT:
		if (*i < n-1)
			*i += 1;
		break;
	case STARTBTN:
		get_xy(&x, &y);
		n = dropdownlist(items, n, *i, x, y);
		if (n)
			*i = n-1;
		setcurs(x, y);
		return 3;
	default:
		return 0;
	}
	putch('[');
	printstr(items[*i]);
	n = getdropdownwidth(items, n) - strlen(items[*i]);
	putnchars(' ', n);
	printstr("] ");
	return 1;
}

/* startup menu */

#ifdef TWOPLAYER
static int gm_mode(int k, int *pos)
{
	const char *items[2] = {"A-type", "B-type"};
	union val v;
	int i;
	int ret;
	v.integ = getopt_int("", "mode");
	i = 1-!(v.integ & MODE_BTYPE);
	ret = selectitem(items, 2, &i, k);
	if (ret && k) {
		if (i)
			v.integ |= MODE_BTYPE;
		else
			v.integ &= ~MODE_BTYPE;
		setoption("", "mode", v, 0);
	}
	return ret;
}

int startupmenu(int i, int x, int y)
{
	union val v;
	const char *menu[6] = {
		"1-Player Game",
		"2-Player Game",
		"Mode",
		"-------------",
		"Options",
		"Highscores"
	};
	menuhandler handlers[6] = {NULL, NULL, gm_mode};
	i = openmenu(menu, 6, i, x, y, handlers);
	if (i==1 || i==2) {
		v.integ = i | getopt_int("", "mode") & MODE_BTYPE;
		setoption("", "mode", v, 0);
	}
	return i;
}
#endif /* TWOPLAYER */
#endif /* !NO_MENU */

/* input setup menu */

#ifdef JOYSTICK
static const char *inp_devkey(const char **sect_name)
{
	const char *s = "";
	const char *dev;
	if (player_)
		s = (player_==1) ? "player1" : "player2";
	dev = getopt_str(s, "input");
	if (sect_name)
		*sect_name = s;
	return dev;
}

static int inp_devhandler(int keypr, int *pos)
{
	const char *sect_name;
	const char *dev = inp_devkey(&sect_name);
	const char *dev2;
	int i = 0;
	union val v;
	if (dev) {
		do i++;
		while ((dev2 = getinputdev_str(i)) && strcmp(dev, dev2));
		if (!dev2)
			i = 0;
		else if (player_) {
			v.integ = inputdevs_player[getinputdev(dev)];
			if (v.integ && v.integ != player_)
				i = 0;
		}
	}
	switch (keypr) {
	case 0:
		break;
	case MVLEFT:
		i--; break;
	case MVRIGHT:
		i++; break;
	case MVDOWN:
	case '\t':
		*pos = 2;
		return 2;
	default:
		return 0;
	}
	dev = getinputdev_str(i);
	if (!dev)
		return 1;
	printmenuitem_options(inputdevs_str, i);
	if (!i)
		unsetoption(sect_name, "input");
	else {
		strncpy(v.str, dev, 4);
		setoption(sect_name, "input", v, 1);
	}
#ifdef TWOPLAYER
	if (player_) {
		i = getinputdev(dev);
		if (i)
			inputdevs_player[i] = player_;
		else {
			inputdevs_player[0] = 0;
			if (player_==2) {
				player_ = 1;
				if (getinputdev(inp_devkey(NULL)))
					inputdevs_player[0] = 2;
				player_ = 2;
			}
		} 
	}
#endif
	if (keypr) {
		*pos = 1;
		return 2;
	}
	return 1;
}
#endif /* JOYSTICK */

static int inp_keypr(int i)
{
	const char keypr[11] = {MVUP, MVDOWN, MVLEFT, MVRIGHT, A_BTN, B_BTN,
				MVUP, A_BTN, B_BTN, HARDDROP, MVDOWN};
	int k = keypr[i];
	if (player_)
		k |= player_==1 ? PLAYER_1 : PLAYER_2;
	if (i >= 6)
		k |= IN_GAME;
	return k;
}

static void inp_printkeys(int dev, int x, int y)
{
	const char *k;
	int i;
	x += 6;
	setcurs(x, y);
	for (i = 0; i < 11; i++) {
		k = getkeyfor_str(dev, inp_keypr(i));
		printstr(k);
		putnchars(' ', 6-strlen(k));
		if (i == 5) {
			x += 17;
			setcurs(x, y);
		} else
			newln(x);
	}
}

#ifdef JOYSTICK
static int inp_devlist(int *dev, int x, int y)
{
	const char *device = "Device";
	menuhandler handler = inp_devhandler;
	int ret = 0;
	drawmenu(&device, 1, 0, x, y, &handler);
	while (1) {
		*dev = getinputdev(inp_devkey(NULL));
		if (ret == 3)
			break;
		inp_printkeys(*dev, x, y+1);
		ret = openmenu(&device, 1, 0, x, y, &handler);
		if (ret < 2)
			return 0;
	}
	setcurs(x, y);
	printmenuitem(device, 0);
	newln(x);
	return 1;
}
#endif

static int inp_setcurs(int x, int y, int i)
{
	if (i >= 6) {
		x += 12;
		i -= 6;
	}
	setcurs(x, y+i);
	return x;
}

void inputsetup_menu(int pl, int x, int y)
{
	const char *menu[11] = {
		" Up", "Dwn", "Lft", "Rgt", "  A", "  B",
		" Rotate ", " Rot cw ", NULL, "Harddrop", "Softdrop"
	};
	char rot_acw[9] = " Rot acw";
	int dev = 0;
	int i = 0;
	int xx, k;
#ifdef JOYSTICK
	int devlist = getinputdev_str(1) != NULL;
	if (devlist) {
		i = -1;
		setcurs(x, y);
		newln(x);
		y++;
	}
#endif
	spellword(rot_acw+5);
	menu[8] = rot_acw;
	drawmenu(menu, 6, i, x, y, NULL);
	drawmenu(menu+6, 5, -1, x+12, y, NULL);
#ifdef TWOPLAYER
	player_ = pl;
#endif
	while (1) {
#ifdef JOYSTICK
		if (devlist && i==-1) {
			if (!inp_devlist(&dev, x, y-1))
				return;
			printmenuitem(menu[0], 1);
			i = 0;
		}
#endif
		inp_printkeys(dev, x, y);
		xx = inp_setcurs(x, y, i);
		refreshwin(-1);
		k = setkeymapping(dev, inp_keypr(i));
#ifdef JOYSTICK
		if (devlist && (k==MVUP && !i || k=='\t' && i==10)) {
			printmenuitem(menu[i], 0);
			i = -1;
			continue;
		}
#endif
		if (k==MVDOWN && i==10) {
			printmenuitem(menu[10], 0);
			inp_setcurs(x, y, i=0);
			printmenuitem(menu[0], 1);
			continue;
		}
		switch (handle_menuitem_2cols(menu, 11, &i, 6,
					      xx, y+i%6, NULL, k)) {
		case 0:
		case 2:
			return;
		case 3:
			inp_setcurs(x, y, i);
			printmenuitem(menu[i], 1);
		}
	}
}

#ifndef NO_MENU
void inputsetup_box(int pl, int x, int y)
{
	int h = 8;
#ifdef JOYSTICK
	if (getinputdev_str(1))
		h++;
#endif
	clearbox(x-1, y, 0, h+1);
	if (h==8) {
		newln(x+2); newln(x+2);
		printstr("Press ENTER to return");
	}
	setcurs(x, y);
	printmenuitem("Input Setup", 1);
#ifdef TWOPLAYER
	if (!pl)
		printstr(" single");
	printstr(" player");
	if (pl)
		putch(pl+'0');
#endif
	drawbox(x, y+1, 31, h, NULL);
	inputsetup_menu(pl, x+1, y+2);
	if (h==8)
		h += 2;
	clearbox(x, y, 0, h+1);
}

static int rarrow_menuitem(int k, int *p)
{
	printstr("\b->");
	return k==MVRIGHT ? 2 : 0;
}

#ifdef JOYSTICK
static int inputsetup_menuitem(int k, int *pos)
{
	int i = 0;
	if (!getinputdev_str(1)) {
		i = rarrow_menuitem(k, pos);
		cleartoeol();
		return i;
	}
	if (inp_devhandler(k, &i)) {
		if (k && i!=1)
			movefwd(strlen(inputdevs_str));
		printstr(" ->");
		if (i==1)
			return 3;
		return k <= MVRIGHT;
	}
	return 0;
}
#else
#define inputsetup_menuitem rarrow_menuitem
#endif

/* game menus */

static int gm_level(int k, int *pos)
{
	struct player *p = game->player+(player_==2);
	int i;
	if (k==MVLEFT)
		k = -1;
	else if (k==MVRIGHT)
		k = 1;
	else if (k >= '0' && k <= '9') {
		p->startlevel = 0;
		k -= '0';
	} else if (k)
		return 0;
	i = p->startlevel+k;
	if (i >= 0 && i < 10) {
		p->startlevel = i;
		printmenuitem_options("0 1 2 3 4 5 6 7 8 9", i);
	}
	return 1;
}

static int gm_height(int k, int *pos)
{
	struct player *p = game->player+(player_==2);
	int i;
	if (k==MVLEFT)
		k = -1;
	else if (k==MVRIGHT)
		k = 1;
	else if (k >= '0' && k <= '5') {
		p->height = 0;
		k -= '0';
	} else if (k)
		return 0;
	i = p->height+k;
	if (i >= 0 && i <= 5) {
		p->height = i;
		printmenuitem_options("0 1 2 3 4 5", i);
	}
	return 1;
}

static int gm_rotate(int k, int *pos)
{
	char s[25] = "clockwise anticlockw.";
	struct player *p = game->player+(player_==2);
	spellword(s+10);
	if (k==MVLEFT)
		p->rotationsys |= ROT_CLOCKWISE;
	else if (k==MVRIGHT)
		p->rotationsys &= ~ROT_CLOCKWISE;
	else if (k)
		return 0;
	printmenuitem_options(s, !(p->rotationsys & ROT_CLOCKWISE));
	return 1;
}

static int gm_orient(int k, int *pos)
{
	struct player *p = game->player+(player_==2);
	if (k==MVLEFT)
		p->rotationsys |= ROT_LEFTHAND;
	else if (k==MVRIGHT)
		p->rotationsys &= ~ROT_LEFTHAND;
	else if (k)
		return 0;
	printmenuitem_options("left right -handed", !(p->rotationsys &
						      ROT_LEFTHAND));
	return 1;
}

static int gm_softdrop(int k, int *pos)
{
	const char steps[5] = {1, 2, 3, 5, 20};
	int i;
	for (i = 0; i < 4; i++)
		if (steps[i] >= softdrop_speed)
			break;
	if (k==MVLEFT) {
		if (!i)
			return 1;
		i--;
	} else if (k==MVRIGHT) {
		if (i==4)
			return 1;
		i++;
	} else if (k)
		return 0;
	if (k)
		softdrop_speed = steps[i];
	printmenuitem_options("1 2 3 5 20", i);
	return 1;
}

#ifndef TWOPLAYER
#define gamemenu_2p(menu, x, y, handlers) 0
#else
static int gamemenu_2p(const char **menu, int x, int y, menuhandler *handlers)
{
	int yy = y;
	int keypr;
	int i1 = 0;
	int i2 = 0;
	int *i = &i1;
begin:	draw_2p_menu_decor(1, x-1, y);
	draw_2p_menu_decor(2, x-1, y+GAMEMENU_LENGTH);
	x += 2;
back:	player_ = 1;
	drawmenu(menu, GAMEMENU_LENGTH-1, i1, x, y, handlers);
	player_ = 2;
	drawmenu(menu, GAMEMENU_LENGTH-1, i2, x, y+GAMEMENU_LENGTH, handlers);
	while (1) {
		setcurs(x, yy+*i);
		refreshwin(-1);
		keypr = getkeypress_block(0);
		if (keypr==STARTBTN && i1!=3 && i2==3)
			keypr |= PLAYER_2;
		if (!(keypr & PLAYER_2)) {
			player_ = 1;
			i = &i1;
			yy = y;
		} else {
			player_ = 2;
			i = &i2;
			yy = y+GAMEMENU_LENGTH;
		}
		setcurs(x, yy+*i);
		switch (handle_menuitem(menu, GAMEMENU_LENGTH-1, i,
					x, yy+*i, handlers, keypr)) {
		case 0:
			return 0;
		case 2:
			if (*i==3) {
				x -= 2;
				clearbox(0, y+9, 0, 4);
				inputsetup_box(player_, x, y);
				goto begin;
			}
			return 1;
		case 3:
			goto back;
		}
	}
}
#endif /* TWOPLAYER */

int gamemenu(int i, int x, int y)
{
	const char *menu[GAMEMENU_LENGTH] = {
		"Level",
		"Height",
		"-",
		"Input Setup",
		"Rotate",
		"Rotation Sys",
		"Softdrop Speed"
#ifndef TWOPLAYER
		,"-",
		"Options",
		"Highscores"
#endif
	};
	menuhandler handlers[GAMEMENU_LENGTH] = {
		gm_level, gm_height, NULL,
		inputsetup_menuitem, gm_rotate, gm_orient,
		gm_softdrop
#ifndef TWOPLAYER
		,NULL, rarrow_menuitem, rarrow_menuitem
#endif
	};
	struct player *p = &player1;
	const char *s = "";
	union val v;
#ifdef SOCKET
	if (socket_fd > -1)
		i = gamemenu_socket(menu, i, x, y, handlers);
	else
#endif
	if (TWOPLAYER_MODE)
		i = gamemenu_2p(menu, x, y, handlers);
	else
		i = openmenu(menu, GAMEMENU_LENGTH, i, x, y, handlers);
	if (i==4)
		return 4;
	if (!i)
		return 0;
#ifndef TWOPLAYER
	if (i >= GAMEMENU_LENGTH-2)
		return i;
#endif
	if (TWOPLAYER_MODE && socket_fd == -1)
		s = "player1";
	while (1) {
		v.integ = p->startlevel;
		setoption(s, "level", v, 0);
		v.integ = p->height;
		setoption(s, "height", v, 0);
		v.integ = p->rotationsys;
		setoption(s, "rotate", v, 0);
		if (!TWOPLAYER_MODE) {
			v.integ = softdrop_speed;
			setoption("", "softdrop", v, 0);
			break;
		}
		if (p == &player2 || socket_fd > -1)
			break;
		p = &player2;
		s = "player2";
	}
	return 1;
}

/* options menu */

#ifdef TWOPLAYER
static int op_inputsetup_select(int x, int y)
{
	const char *items[3] = {
		"single player",
		"  player1",
		"  player2"
	};
	return dropdownlist(items, 3, 0, x+14, y);
}
#else
#define op_inputsetup_select(x, y) 1
#endif

static void save_bgdot()
{
	union val v;
	v.str[0] = bgdot;
	v.str[1] = '\0';
	setoption("term", "bgdot", v, 1);
}

static int op_bgdot(int k, int *pos)
{
	const char *items[3] = {"dot", "bullet", "blank"};
	int n = 3;
	int i = 0;
	int j;
	if (isprintable(k)) {
		if (k == '0')
			bgdot = ' ';
		else if (k == '~')
			bgdot = BULLET;
		else
			bgdot = k;
		save_bgdot();
		k = 0;
	}
	if (_ASCII) {
		items[1] = items[2];
		n--;
	} else if (bgdot == BULLET)
		i = 1;
	if (bgdot == ' ')
		i = n-1;
	j = i;
	k = selectitem(items, n, &i, k);
	if (i != j) {
		if (i == 0)
			bgdot = '.';
		else if (i == n-1)
			bgdot = ' ';
		else
			bgdot = BULLET;
		save_bgdot();
	}
	if (k != 1)
		return k;
	putch(bgdot);
	putch(' ');
	return 1;
}

static int op_tcolor(int k, int *pos)
{
	char *c = tetrom_colors + *pos;
	if (k >= '1' && k <= '7')
		*c = k-'0';
	else if (k)
		return 0;
	setcolorpair(*c);
	putch(' ');
	putch(*c+'0');
	putch(' ');
	setattr_normal();
	return 1;
}

static int op_tcolor_gt4(int k, int *pos)
{
	int i = *pos+4;
	return op_tcolor(k, &i);
}

static void tetrom_colors_menu(int x, int y)
{
	const char *menu[7] = {"I", "J", "L", "O", "S", "T", "Z"};
	menuhandler handlers[7] = {
		op_tcolor, op_tcolor, op_tcolor, op_tcolor,
		op_tcolor_gt4, op_tcolor_gt4, op_tcolor_gt4
	};
	int i = 0;
	if (y+6 >= term_height)
		y -= 7;
	drawbox(x, y, 19, 6, NULL);
	x++;
	y++;
	drawmenu(menu, 4, 0, x, y, handlers);
	drawmenu(menu+4, 3, -1, x+9, y, handlers+4);
	while (1) {
		setcurs(x, y+i%4);
		refreshwin(-1);
		switch (handle_menuitem_2cols(menu, 7, &i, 4, x, y+i%4,
				handlers, getkeypress_block(SINGLE_PL))) {
		case 0:
		case 2:
			if (i >= 4)
				x -= 9;
			clearbox(x-1, y-1, 19, 6);
			return;
		case 3:
			x += i>=4 ? 9 : -9;
			setcurs(x, y+i%4);
			printmenuitem(menu[i], 1);
		}
	}
}

void optionsmenu(const char **items, int n, menuhandler f, int x, int y)
{
	const char *menu[8] = {"Input Setup", "-"};
	char tetr_colors[18] = "Tetromino Colours";
	menuhandler handlers[8] = {rarrow_menuitem};
	int i = 1;
	memcpy(&menu[2], items, n*sizeof(char *));
	while (i <= n)
		handlers[++i] = f;
	menu[++i] = "Block Style";
	handlers[i] = f;
	menu[++i] = "Board BG";
	handlers[i] = op_bgdot;
	spellword(tetr_colors+10);
	menu[++i] = tetr_colors;
	handlers[i] = rarrow_menuitem;
	n += 5;
	i = 0;
	while (i = openmenu(menu, n, i, x, y, handlers)) {
		i--;
		if (!i)
			f(op_inputsetup_select(x, y), &i);
		else if (i == n-1)
			tetrom_colors_menu(x, y+i+1);
		else if (!handlers[i](STARTBTN, &i))
			break;
	}
	clearbox(x, y, 0, n);
}

void show_hiscorelist(int x, int y)
{
	char buf[320];
	char *p;
	int n;
	setcurs(x, y);
	printmenuitem("Highscores", 1); newln(x);
	setcolorpair(MAGENTA_FG);
	printstr("    Name      Score  Lvl  Lines"); newln(x);
	setattr_normal();
	if (!readhiscores(NULL)) {
		printstr("    No saved scores");
		newln(x);
	} else {
		n = gethiscorelist(buf);
		p = buf;
		while (n) {
			p[30] = '\0';
			printstr(p);
			newln(x);
			p += 32;
			n--;
		}
	}
	refreshwin(-1);
	if (getkeypress_block(SINGLE_PL) == 'q')
		exit(0);
}
#endif /* !NO_MENU */

void show_hiscorelist5(int x, int y, int i)
{
	char s[20];
	const struct hiscore *hs, *end;
	int n;
	if (!hiscores[0].score && !readhiscores(NULL))
		return;
	setcurs(x, y);
	hs = hiscores;
	i -= 5;
	if (i < 0)
		i = 0;
	else
		hs += i;
	end = hs+5;
	while (hs->score && hs != end) {
		n = sprintf(s, "%2d. %s", i+1, gethiscorename(i, s+12));
		memset(s+n, ' ', 12-n);
		sprintf(s+12, "%7d", hs->score);
		printstr(s);
		newln(x);
		hs++;
		i++;
	}
}

static void printtextbox(const char *text, int pos)
{
	int underscore;
	if (_WHITE_BG)
		underscore = term_flags & (ASCII | GNOME_TERM) || !_XTERM;
	else
		underscore = term_flags & (MONOCHROME | TT_BLOCKS);
	while (*text) {
		setattr_normal();
		if (!pos)
			setattr_standout();
		else {
			if (_WHITE_BG) {
				if (!underscore || term_flags & GNOME_TERM)
					setattr_underline();
			} else
				setcolorpair(WHITE_ON_BLUE);
		}
		if (*text == ' ' && underscore)
			putch('_');
		else
			putch(*text);
		text++;
		pos--;
	}
	setattr_normal();
}

static void printsavebutton(int sel)
{
	if (_MONOCHROME) {
		if (sel)
			setattr_standout();
		else
			setattr_bold();
	} else if (sel) {
		setcolorpair(YELLOW_ON_GREEN);
	} else {
		setcolorpair(4);
		putch('[');
		setcolorpair(YELLOW_ON_BLUE);
		printstr(" Save ");
		setcolorpair(4);
		putch(']');
		goto out;
	}
	printstr("[ Save ]");
out:	setattr_normal();
}

static int hiscore_editname(char *name, int pos, int k)
{
	const char chars[18] = "Z.!?-0123456789 A";
	const char lat1_chars[5] = "";
	const char lat1_chars_small[5] = "";
	int c, i;
	char *p;
	if (k == MVLEFT && pos)
		pos--;
	else if (k == '\b' && pos) {
		pos--;
		memmove(name+pos, name+pos+1, 6-pos);
		name[6] = ' ';
	} else if (pos == 7)
		return pos;
	else if (k == MVRIGHT)
		pos++;
	else if (k == '\t')
		pos = 7;
	else if (k == 0x7F) {
		memmove(name+pos, name+pos+1, 6-pos);
		name[6] = ' ';
	} else {
		c = 0;
		i = 0;
		if (k >= 'A' && k <= 'Z' || strchr(chars, k) ||
					    strchr(lat1_chars, k))
			c = k;
		else if (k >= 'a' && k <= 'z')
			c = k-32;
		else if (p = strchr(lat1_chars_small, k))
			c = lat1_chars[p-lat1_chars_small];
		else if (k == MVUP || k == A_BTN || k == B_BTN) {
			c = name[pos];
			i = (k==B_BTN) ? -1 : 1;
			if (c > 'A'-(i>0) && c < 'Z'+(i<0))
				c += i;
			else
				c = *(strchr(chars, c)+i);
		}
		if (c) {
			name[pos] = c;
			if (!i)
				pos++;
		}
	}
	return pos;
}

static int hiscore_entername_menu(char *name, const char **menu, int x, int y)
{
	int pos = 0;
	int i = 0;
	int k;
	while (1) {
		drawmenu(menu, 2, i-2, x, y+2, NULL);
textbox:	setcurs(x, y);
		printtextbox(name, pos);
		movefwd(3);
		printsavebutton(!i && pos==7);
		if (!i)
			x += pos;
		setcurs(x, y+i);
		refreshwin(-1);
		k = getkeypress_block(SINGLE_PL) & 0xFF;
		if (k == ESC)
			return 0;
		if (!i) {
			x -= pos;
			switch (k) {
			case STARTBTN:
				return 1;
			case '\t':
				if (pos < 7)
					break;
			case MVDOWN:
				i = 2;
				continue;
			}
			pos = hiscore_editname(name, pos, k);
			goto textbox;
		}
		switch (k) {
		case STARTBTN:
		case A_BTN:
			if (i==2)
				return 2;
		case 'q':
			exit(0);
		case MVUP:
			i = i==2 ? 0 : 2;
			break;
		case '\t':
			i = i==2 ? 3 : 0;
			break;
		case MVDOWN:
			if (i < 3)
				i++;
		}
		if (!i && pos==7) {
			pos = 6;
			while (pos && name[pos-1]==' ')
				pos--;
		}
	}
}

static int playagain_menu(const char **menu, int x, int y)
{
	int ret = openmenu(menu, 2, 0, x, y, NULL);
	if (ret==2)
		exit(0);
	return ret;
}

static int hiscorebox(const char **menu, int x, int y)
{
	int i = 0;
	while (i < 10 && hiscores[i].score >= player1.score)
		i++;
	drawbox(x, y, 24, 11, "HIGHSCORES");
	x += 2;
	show_hiscorelist5(x, y+2, i);
	if (playagain_menu(menu, x+2, y+8)) {
		clearbox(x+21, y, 0, 11);
		return 1;
	}
	return 0;
}

static int hiscore_congrats(const char **menu)
{
	char name[8];
	int y = _HEIGHT_24L ? 7 : 3;
	int i;
	setwcurs(0, 9, y);
	drawbox(9, y, 26, 9, "CONGRATULATIONS!");
	setcurs(11, y+2);
	printstr("You have a highscore!"); newln(11);
	printstr("Please enter your name");
	while (1) {
		strcpy(name, "       ");
		switch (hiscore_entername_menu(name, menu, 13, y+4)) {
		case 0:
			return 0;
		case 1:
			i = 6;
			while (name[i] == ' ') {
				name[i] = '\0';
				i--;
			}
			if (savehiscore(name)) {
				clearbox(33, y, 2, 9);
				return hiscorebox(menu, 9, y);
			}
			setcurs(11, y+2);
			printstr("ERROR! Could not save"); newln(11);
			printstr("score to file.        ");
			continue;
		case 2:
			clearbox(32, y, 0, 9);
			return 1;
		}
	}
}

int gameovermenu()
{
	const char *menu[2] = {"Play again", "Exit"};
	readhiscores(NULL);
	if (ishiscore())
		return hiscore_congrats(menu);
	setwcurs(1, 2, 3);
	drawbox(2, 3, 17, 5, "GAME OVER");
	return playagain_menu(menu, 4, 5);
}
